"""
Note that for pytorch lightning you should pass the lightning trainer into WorldInfo


Rank: Global rank of the process in the distributed setting.
Local rank: Rank on this machine

Example:
            |    Node1  |   Node2    |
____________| p1 |  p2  |  p3  |  p4 |
local_rank  | 0  |   1  |  0   |   1 |
rank        | 0  |   1  |  2   |   4 |

"""

from __future__ import annotations

import multiprocessing
import os
from deprecated import deprecated
from multiprocessing.process import BaseProcess


class WorldInfo:
    def __init__(self, trainer=None):
        self.trainer = trainer

    @property
    def global_rank(self):
        if self.trainer is not None:
            return self.trainer.global_rank
        return get_global_rank()

    @property
    def world_size(self):
        if self.trainer is not None:
            return self.trainer.world_size
        return get_world_size()

    @property
    def is_global_zero(self):
        return self.global_rank == 0

    def print_with_rank(self, *args, **kwargs):
        rank = self.global_rank
        world_size = self.world_size
        print(f"Rank {rank:>2d}/{world_size}:", *args, **kwargs)

    def barrier_safe(self):
        if self.trainer is not None:
            return self.trainer.strategy.barrier()
        return barrier_safe()


def get_process_info() -> str:
    process: BaseProcess = multiprocessing.current_process()
    return f"Process {process.pid} {repr(process)}"


def get_world_info() -> tuple[int, int]:
    """
    Return global rank and world size from environment variables. If not set, return 0, 1.
    """
    return get_global_rank(), get_world_size()


def get_world_size() -> int:
    if is_slurm_sbatch():
        return int(os.environ["SLURM_NTASKS"])
    return int(os.environ.get("WORLD_SIZE", 1))


def get_global_rank() -> int:
    """
    In some cases LOCAL_RANK is set, but RANK is unset. Use LOCAL_RANK in that case.
    RANK: global rank of the process in the distributed setting, across all nodes.
    LOCAL_RANK: rank on this machine / node.

    in slurm sbatch scripts, we need to use the slurm env variables instead.
    """
    if is_slurm_sbatch():
        return int(os.environ["SLURM_PROCID"])

    if "RANK" in os.environ:
        rank = int(os.environ["RANK"])
    else:
        rank = int(os.environ.get("LOCAL_RANK", 0))
    return rank


@deprecated(reason="Use get_global_rank instead")
def get_rank():
    return get_global_rank()


def is_slurm_sbatch():
    slurm_job_name = os.environ.get("SLURM_JOB_NAME")
    if slurm_job_name is None:
        # not in slurm job
        return False
    if slurm_job_name == "bash":
        # in foreground slurm job
        return False
    if "SLURM_NTASKS" not in os.environ:
        # special setup with e.g. background slurm job that runs a jupyter frontend
        return False
    # actual background slurm job
    return True


def print_main(*args, **kwargs):
    if is_main_process():
        print(*args, **kwargs)


def print_with_rank(*args, **kwargs):
    rank = get_rank()
    world_size = get_world_size()
    print(f"Rank {rank:>2d}/{world_size}:", *args, **kwargs)


def is_main_process():
    rank, _world_size = get_world_info()
    return rank == 0


def barrier_safe():
    """Barrier only if in a distributed torch run. Does not fail if torch package is missing."""
    if is_distributed():
        from torch import distributed as dist

        dist.barrier()


def is_distributed():
    return "WORLD_SIZE" in os.environ and get_world_size() > 1



from tempfile import NamedTemporaryFile
from typing import Tuple

import numpy as np
from loguru import logger
from pydub import AudioSegment
from scipy.io import wavfile
from scipy.signal import spectrogram

from packg import Const
from packg.typext import PathType


class AudioConvertC(Const):
    NONE = "none"
    MONO = "mono"
    STEREO = "stereo"


def load_audio_file(
    input_file: PathType, verbose: bool = False, convert=AudioConvertC.STEREO
) -> Tuple[int, np.ndarray]:
    """
    Load audio file and return frequency and data.

    Args:
        input_file:
        verbose: print debug info
        convert: convert to either stereo (shape (N, 2)) or mono (shape (N,)) or leave as is.

    Returns:
        tuple of frequency int and data numpy array of shape (N, 2) (stereo) or (N,) (mono)

    """
    file_format = input_file.name.split(".")[-1]
    if verbose:
        logger.info(f"Reading file {input_file} detected format {file_format}")
    audio = AudioSegment.from_file(input_file, format=file_format)

    # convert to wav
    with NamedTemporaryFile(suffix=".wav") as tmpfile:
        audio.export(tmpfile, format="wav")
        frequency, data = wavfile.read(tmpfile)

    # convert to stereo or mono as needed
    if convert == AudioConvertC.STEREO and len(data.shape) == 1:
        data = np.stack([data, data], axis=-1)  # mono to stereo
    elif convert == AudioConvertC.MONO and len(data.shape) == 2:
        data = data.mean(-1)  # stereo to mono

    if verbose:
        logger.info(f"Got frequency {frequency} shape {data.shape} after conversion '{convert}'")

    return frequency, data


def check_audio_quality(frequency, data, eps=1e-6) -> bool:
    """
    logic: a good quality music file will have audio at 20khz. a bad one will not.
    """
    # dmin, dmax = data.min(), data.max()
    # dnorm = (data - dmin) / (dmax - dmin)
    # print_ndarray_lovely(data)  # raw mono audio
    # example data shape 11,796,960 freq 48000 => 246 seconds

    freqs, time, spec = spectrogram(data, fs=frequency, nperseg=256, noverlap=128)

    # print_ndarray_lovely(freqs)  # frequency bins
    # select all bins within 19 and 21 khz
    freqs_to_use = np.less_equal(19000, freqs) & np.less_equal(freqs, 21000)
    # example: 129 equally spaced values from 0 ... 24000

    # print_ndarray_lovely(time)  # time bins
    # print_ndarray_lovely(spec)  # spec shape (n_freq, n_time)
    smin, smax = spec.min(), spec.max()
    snorm = (spec - smin) / (smax - smin)

    srel = snorm[freqs_to_use]
    return srel.max() > eps
    # now we want to normalize with something between mean (1-norm) and max (inf-norm)
    # srel_norm_freq = np.linalg.norm(srel, ord=np.inf, axis=0)
    # srel_norm_all = np.linalg.norm(srel_norm_freq, ord=np.inf, axis=0)
    # return srel_norm_all
